/*
 * time_cvt.c - convert text times <-> seconds
 */
#include <SANE.h>
#include "RevRdist.h"
#include <Script.h>
#include <Packages.h>

OSErr	str2secs(Str255, Longint *, Str255);



/*
 *=========================================================================
 * str2secs(str, secp, outstr) - convert string to binary seconds
 *				and/or binary seconds to string
 * entry:	str = string to convert
 *			secp = ptr to long to hold binary seconds
 *			outstr = ptr to string to hold back conversion
 *			If str is not empty, it is converted to seconds into *secp.
 *			This value is then converted back to text in outstr.
 *			If str is null or empty, only the back conversion from
 *			secp to outstr is performed.
 * exit:	Returns 0 if conversions successful,
 *			OSErr on system errors
 *			1 if text is not a valid time string, secp contains detailed
 *			error code
 *=========================================================================
 */


OSErr
str2secs(Str255 str, Longint *secp, Str255 outstr)
{
	OSErr				error;
	Handle				itl0;			/* handle to itl0 */
	Itl4Handle			itl4;			/* number formatting resource */
	NumberParts			roman;			/* English number parts */
	NumberParts			parts;			/* current script number parts */
	NumberPartsPtr		pp;				/* ptr to parts */
	NumFormatString		fs;				/* internal format for number fmt */
	Str255				st;				/* temp string */
	SignedByte			hs;				/* save handle state */
	extended			x;
	long double			d;
	long				days, secs, daypart;
	DateCacheRecord		dc;
	long				inlen, outlen;
	LongDateRec			dt;
	union LongDateCvt 	cd;
	LongDateTime		td;

	error = 0;
	if (secp == NULL)
		return paramErr;
	/*
	 * convert English ###### format to internal form
	 */
	secs = *secp;
	itl4 = (Itl4Handle) GetResource('itl4', 0);
	if (itl4 == 0)
		goto syserr;
	hs = HGetState((Handle)itl4);
	HLock ((Handle)itl4);
	pp = (NumberPartsPtr)((char *)(*itl4) + (*itl4)->defPartsOffset);
	roman = *pp;
	HSetState((Handle)itl4, hs);
	error = Str2Format ("\p######", &roman, &fs);
	if (error)
		goto fmterr;
	/*
	 * Now get the parts table for the current script
	 */
	itl4 = (Itl4Handle) IUGetIntl(4);
	if (itl4 == 0)
		goto syserr;
	hs = HGetState((Handle)itl4);
	HLock ((Handle)itl4);
	pp = (NumberPartsPtr)((char *)(*itl4) + (*itl4)->defPartsOffset);
	parts = *pp;
	HSetState((Handle)itl4, hs);
	error = InitDateCache(&dc);
	if (error)
		goto syserr;
	/*
	 * If there is no input string, skip conversion of it to seconds
	 */
	if (str[0] == 0)
		str = NULL;
	if (str == NULL)
		goto convert;
	/*
	 * Convert string to seconds.
	 * First see if string is a simple number
	 */
	error = FormatStr2X(str, &fs, &parts, &x);
	if (error == 0)
	{
#if	!__option(mc68881) && __option(native_fp)
		d = x;
#else
		x80tox96(&x, &d);
#endif
		secs = d;
		if (secs >= 0)
		{
			/*
			 * Numbers less than 24 are days.
			 * Numbers between 24 and 60 are hours.
			 * Numbers greater than 60 are seconds.
			 */
			if (secs < 24)
				secs *= (60 * 60 * 24l);
			else if (secs < 60)
				secs *= (60 * 60);
			goto convert;
		}
		else
			error = fSpuriousChars;
	}
	/*
	 * Did not convert as simple number
	 */
	if (error != fSpuriousChars && error != fBestGuess)
		goto fmterr;
	/*
	 * Try to convert as just time
	 */
	outlen = 0;
	inlen = str[0];
	ZERO(dt);
	error = String2Time ((Ptr)str+1, inlen, &dc, &outlen, &dt);
	if (error == 0)
	{
		dt.ld.year = 1904;
		dt.ld.month = 1;
		dt.ld.day = 1;
		LongDate2Secs (&dt, &td);
		cd.c = td;
		secs = cd.hl.lLow;
		goto convert;
	}
	if (error == cantReadUtilities)
		goto fmterr;
	/*
	 * That didn't work, try converting as days followed by time
	 */
	outlen = inlen;
	error = 0;
	/*
	 * Find longest string which could be the days field
	 */
	d = 0;
	for (inlen = 1; !error && inlen < outlen; inlen++)
	{
		str[0] = inlen;
		error = FormatStr2X(str, &fs, &parts, &x);
		if (error == 0)
		{
	#if	!__option(mc68881) && __option(native_fp)
			d = x;
	#else
			x80tox96(&x, &d);
	#endif
		}
	}
	days = d;
	str[0] = outlen;
	if (error != fSpuriousChars && error != fBestGuess)
		goto fmterr;
	/*
	 * Convert rest as time
	 */
	error = String2Time ((Ptr)str+inlen, outlen - inlen + 1, &dc, &outlen, &dt);
	if (error & fatalDateTime)
		goto fmterr;

	dt.ld.year = 1904;
	dt.ld.month = 1;
	dt.ld.day = 1;
	LongDate2Secs (&dt, &td);
	cd.c = td;
	secs = cd.hl.lLow + days * (60 * 60 * 24l);

	/*
	 * Now, convert seconds back to canonical form
	 */
convert:
	if (secs < 0)
		return paramErr;
	if (outstr)
	{
		days = secs / (60 * 60 * 24l);
		daypart = secs - days * (60 * 60 * 24l);
		cd.hl.lHigh = 0;
		cd.hl.lLow = daypart;
		LongSecs2Date (&cd.c, &dt);
		itl0 = IUGetIntl(0);
		if (itl0 == 0)
			goto syserr;
		error = HandToHand (&itl0);
		if (error || itl0 == 0)
			goto syserr;
		(*(Intl0Hndl)itl0)->timeCycle = 0;
		(*(Intl0Hndl)itl0)->timeFmt |= secLeadingZ|minLeadingZ|hrLeadingZ;
		(*(Intl0Hndl)itl0)->time1Suff = 0;
		IULTimeString (&cd.c,
					   (dt.ld.second || secs < 10 * 60) ? true : false,
					   st, itl0);
		DisposHandle (itl0);
		if (days)
		{
			d = days;
#if	!__option(mc68881) && __option(native_fp)
			x = d;
#else
			x96tox80(&d, &x);
#endif
			error = FormatX2Str (x, &fs, pp, outstr);
			if (error)
				goto fmterr;
			if (st[0])
			{
				outstr[++outstr[0]] = ' ';
				outstr[++outstr[0]] = ' ';
			}
		}
		else
			outstr[0] = 0;
		/*
		 * Concatenate full days and time string
		 */
		inlen = st[0];
		BlockMove ((Ptr) st + 1, (Ptr) outstr + outstr[0] + 1, inlen);
		outstr[0] += inlen;
	}
	if (str)
		*secp = secs;

	return 0;

	/*
	 * Here on unexpected conversion errors
	 */
fmterr:
	if (str)
		*secp = error;
	return 1;

	/*
	 * Here on system errors
	 */
syserr:
	if (error == 0)
	{
		error = ResError();
		if (error == 0)
			error = MemError();
		if (error == 0)
			error = 1;
		return error;
	}
}
